import yaml

from charmhelpers.core import host, hookenv, unitdata
from charmhelpers.core.templating import render
from charmhelpers import context
from charms.reactive import (
    when, when_not, when_any, set_state, remove_state
)
from charms.reactive.helpers import any_file_changed, data_changed
from charms.layer import snap


SNAP_NAME = 'prometheus-alertmanager'
SVC_NAME = 'snap.prometheus-alertmanager.alertmanager'
ALERTMANAGER_YML = '/var/snap/prometheus-alertmanager/current/alertmanager.yml'
ALERTMANAGER_DEF = '/var/snap/prometheus-alertmanager/current/daemon_arguments'
CUSTOM_TEMPLATE = '/var/snap/prometheus-alertmanager/common/customer.tmpl'
ALERTMANAGER_YML_TMPL = 'alertmanager.yml.j2'
ALERTMANAGER_DEF_TMPL = 'daemon_arguments.j2'
CUSTOM_TEMPLATE_TMPL = 'customer.tmpl.j2'


def templates_changed(tmpl_list):
    return any_file_changed(['templates/{}'.format(x) for x in tmpl_list])


def get_peers():
    peers = []
    if not context.Relations().peer:
        return False
    for _, data in context.Relations().peer.items():
        peers.append('{}:9094'.format(data.get('private-address')))
    return sorted(peers)


@when_not('alertmanager.installed')
def install_packages():
    hookenv.status_set('maintenance', 'Installing software')
    config = hookenv.config()
    channel = config.get('snap_channel', 'stable')
    snap.install(SNAP_NAME, channel=channel, force_dangerous=False)
    set_state('alertmanager.installed')
    set_state('alertmanager.do-check-reconfig')


def runtime_args(key=None, value=None, update=True):
    kv = unitdata.kv()
    args = kv.get('runtime_args', {})
    if key:
        if update:
            # Ensure we wipe old format options
            if key.replace('--', '-') in args:
                args.pop(key.replace('--', '-'))
            args.update({key: value})
            kv.set('runtime_args', args)
        else:
            if not args.get(key, []):
                cur = set([])
            else:
                cur = set(args.get(key))
            new = list(cur.union((value,)))
            args.update({key: sorted(new)})
            kv.set('runtime_args', args)

    args_list = []
    for k, v in args.items():
        if not v:
            continue
        if isinstance(v, list):
            for i in v:
                args_list.append('{} {}'.format(k, i))
        else:
            args_list.append('{} {}'.format(k, v))

    # sorted list is needed to avoid data_changed() false-positives
    return sorted(args_list)


def check_ports(new_port):
    kv = unitdata.kv()
    if kv.get('alertmanager.port') != new_port:
        hookenv.open_port(new_port)
        if kv.get('alertmanager.port'):  # Dont try to close non existing ports
            hookenv.close_port(kv.get('alertmanager.port'))
        kv.set('alertmanager.port', new_port)


def validate_config():
    return yaml.safe_load(open(ALERTMANAGER_YML))


@when('alertmanager.do-reconfig-yml')
def write_alertmanager_config_yml():
    config = hookenv.config()

    # customer.tmpl
    render(source=CUSTOM_TEMPLATE_TMPL,
           target=CUSTOM_TEMPLATE,
           context={'prometheus_context': config['prometheus_context']})

    # alertmanager.yml
    render(source=ALERTMANAGER_YML_TMPL,
           target=ALERTMANAGER_YML,
           context={'config': config}
           )
    validate_config()
    set_state('alertmanager.do-restart')
    remove_state('alertmanager.do-reconfig-yml')


@when('alertmanager.do-reconfig-def')
def write_alertmanager_config_def():
    config = hookenv.config()
    port = config.get('port', '9093')
    check_ports(port)
    runtime_args('--web.listen-address',
                 ':{}'.format(port))
    if config.get('external_url', False):
        vars = {
            'private_address': hookenv.unit_get('private-address'),
            'public_address': hookenv.unit_get('public-address'),
            # alertmanager default:
            'port': port,
        }
        runtime_args('--web.external-url',
                     config['external_url'].format(**vars))
    if get_peers():
        hookenv.log('Found cluster peers: {}'.format(get_peers()))
        runtime_args('--mesh.listen-address', '0.0.0.0:9094')
        for peer in get_peers():
            runtime_args('--mesh.peer', peer, False)
    else:
        runtime_args('--mesh.listen-address', False)
        runtime_args('--mesh.peer', False)
    args = runtime_args()
    hookenv.log('runtime_args: {}'.format(args))
    if args:
        render(source=ALERTMANAGER_DEF_TMPL,
               target=ALERTMANAGER_DEF,
               context={'args': args},
               )
    set_state('alertmanager.do-restart')
    remove_state('alertmanager.do-reconfig-def')


@when_not('alertmanager.started')
def setup_alertmanager():
    hookenv.status_set('maintenance', 'Installing software')
    install_packages()
    set_state('alertmanager.do-check-reconfig')


@when('alertmanager.installed')
@when_any('alertmanager.started', 'alertmanager.do-check-reconfig')
def check_reconfig_alertmanager():
    config = hookenv.config()
    args = runtime_args()
    if config.changed('snap_channel'):
        remove_state('alertmanager.installed')
        return

    if data_changed('alertmanager.config', config):
        set_state('alertmanager.do-reconfig-yml')
        set_state('alertmanager.do-reconfig-def')

    if data_changed('alertmanager.peers', get_peers()):
        set_state('alertmanager.do-reconfig-def')

    if templates_changed([ALERTMANAGER_YML_TMPL]):
        set_state('alertmanager.do-reconfig-yml')

    if any((
        data_changed('alertmanager.args', args),
        templates_changed([ALERTMANAGER_DEF_TMPL]),
    )):
        set_state('alertmanager.do-reconfig-def')

    remove_state('alertmanager.do-check-reconfig')


@when('alertmanager.do-restart')
def restart_alertmanager():
    if not host.service_running(SVC_NAME):
        hookenv.log('Starting {}...'.format(SVC_NAME))
        host.service_start(SVC_NAME)
    else:
        hookenv.log('Restarting {}, config file changed...'.format(SVC_NAME))
        host.service_restart(SVC_NAME)
    hookenv.status_set('active', 'Ready')
    set_state('alertmanager.started')
    remove_state('alertmanager.do-restart')


# Relations
@when('alertmanager.started')
@when('alertmanager-service.available')
def configure_alertmanager_relation(alert_service):
    config = hookenv.config()
    alert_service.configure(config.get('port'))
